import torch
import numpy as np

__all__ = ['extract_sample']

def extract_sample(n_way, n_support, n_query, datax, datay):
    """
  Picks random sample of size n_support+n_querry, for n_way classes
  Args:
      n_way (int): number of classes in a classification task
      n_support (int): number of labeled examples per class in the support set
      n_query (int): number of labeled examples per class in the query set
      datax (np.array): dataset of images
      datay (np.array): dataset of labels
  Returns:
      (dict) of:
        (torch.Tensor): sample of images. Size (n_way, n_support+n_query, (dim))
        (int): n_way
        (int): n_support
        (int): n_query
    """
    K = np.random.choice(np.unique(datay), n_way, replace=False)
    sample = np.empty([128],dtype=float)
    for cls in K:
        ##pass a df with event id , taxcode, after_image_id, embeddings
        temp_data = datax[datax['new_taxcode']==cls]
        datax_cls = np.array(temp_data['feat_vec'])

        perm = np.random.permutation(datax_cls)
        sample_cls = perm[:(n_support+n_query)]
        for i in range(len(sample_cls)):
            sample = np.vstack((sample,np.array(sample_cls[i],dtype=float)))
    sample = torch.from_numpy(sample)
    return({
        'images': sample[1:],
        'n_way': n_way,
        'n_support': n_support,
        'n_query': n_query
        })



##function for convolutional blocks module

import torch
import torch.nn as nn

__all__ = ['ConvBlock']

class ConvBlock(nn.Module):

    def __init__(self,*args,**kwargs):
        """
        Function to create a convolutional block of different forms
        Inputs:
            depth (int): Number of convolutional layers
            in_channels (int or list): Number of input channels
            out_channels (int or list): Number of output channels
            pooling_kernal (int): Pooling kernal size
            pooling_stride (int): Pooling stride size
            stride (int or list): Stride size
            padding (int or list): Padding size
            kernel_size (int or list): Kernel size
            activation (str): Activation function
            kwargs (dict): Keyword arguments
                conv_type (str): Type of convolutional layer
                pooling_type (str): Type of pooling layer
                dropout (float): Dropout rate
        Returns:
            conv_block (torch.nn.Module): Convolutional block
        """
        super(ConvBlock, self).__init__()

        if 'conv_depth' in kwargs:
            depth = kwargs['conv_depth']
        else:
            depth = 2
        if 'in_channels' in kwargs:
            in_channels = kwargs['in_channels']
        else:
            in_channels = 64
        if 'out_channels' in kwargs:
            out_channels = kwargs['out_channels']
        else:
            out_channels = 64
        if 'kernel_size' in kwargs:
            kernel_size = kwargs['kernel_size']
        else:
            kernel_size = 3
        if 'stride' in kwargs:
            stride = kwargs['stride']
        else:
            stride = 1
        if 'pooling_kernal' in kwargs:
            pooling_kernal = kwargs['pooling_kernal']
        else:
            pooling_kernal = 2
        if 'pooling_stride' in kwargs:
            pooling_stride = kwargs['pooling_stride']
        else:
            pooling_stride = 2
        if 'padding' in kwargs:
            padding = kwargs['padding']
        else:
            padding = 0
        if 'pooling_type' in kwargs:
            self.pool = kwargs['pooling_type']
        else:
            self.pool = 'MaxPool2d'
        if 'dropout' in kwargs:
            self.dropout = kwargs['dropout']
        else:
            self.dropout = 0.0
        
        if 'activation' in kwargs:
            if  kwargs['activation'] == 'relu':
                self.activation = 'ReLU'
            elif kwargs['activation'] == 'leaky_relu':
                self.activation = 'LeakyReLU'
            else:
                activation = kwargs['activation']
        else:
            self.activation = 'ReLU'

        if 'conv_type' in kwargs:
            conv_type = kwargs['conv_type']
        else:
            conv_type = 'Conv2d'
            conv = getattr(nn,conv_type)

        self.convs = nn.Sequential()
        for i in range(depth):
            if isinstance(in_channels,list):
                in_channels = in_channels[i]
            else:
                if i==0:
                    in_channels = in_channels
                else:
                    in_channels = out_channels
            if isinstance(out_channels,list):
                out_channels = out_channels[i]
            else:
                out_channels = out_channels
            if isinstance(stride,list):
                stride = stride[i]
            else:
                stride = stride
            if isinstance(padding,list):
                padding = padding[i]
            else:
                padding = padding
            if isinstance(kernel_size,list):
                kernel_size = kernel_size[i]
            else:
                kernel_size = kernel_size
            #temp_str = 'conv'+str(i+1)
            temp_conv = conv(in_channels=in_channels,out_channels=out_channels,stride=stride,padding=padding,kernel_size=kernel_size)
            #temp_2 = '{}={}'.format(temp_str,temp_conv)
            #exec(temp_2)
            self.convs.add_module(f"conv_{i}",temp_conv)
            self.convs.add_module(f"activation_{i}",getattr(nn,self.activation)())
            self.convs.add_module(f"dropout_{i}",getattr(nn,'Dropout')(self.dropout))

        if 'sample_type' in kwargs:
            sample_type = kwargs['sample_type']
        else:
            sample_type = None
        if sample_type =='up':
            self.convs.add_module(f'up_sample_{i}',getattr(nn,'Upsample')(scale_factor=2,mode='bilinear',align_corners=True))
        elif sample_type =='bottleneck':
            pass
        else:
            self.convs.add_module(f"pool_conv_block_{i}",getattr(nn,self.pool)(kernel_size=pooling_kernal,stride=pooling_stride))

    def forward(self,x):
        x = self.convs(x)
        return x


import torch.nn.functional as F

__all__ = ['dice_loss', 'binary_cross_entropy', 'cross_entropy', 'mse_loss', 'l1_loss']

def dice_loss(input, target):
    """
    Dice loss function
    Formula: 1 - (2 * intersection + smooth) / (union + smooth)
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        dice_loss: torch tensor of shape (1)
    """
    smooth = 1.
    iflat = input.view(-1)
    tflat = target.view(-1)
    intersection = (iflat * tflat).sum()
    return 1 - ((2. * intersection + smooth) /
                (iflat.sum() + tflat.sum() + smooth))

def binary_cross_entropy(input, target):
    """
    Binary cross entropy loss function
    Formula: -sum(y * log(x) + (1 - y) * log(1 - x))
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        binary_cross_entropy_loss: torch tensor of shape (1)
    """
    return F.binary_cross_entropy(input, target)

def cross_entropy(input, target):
    """
    Cross entropy loss function
    Formula: -sum(y * log(x))
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        cross_entropy_loss: torch tensor of shape (1)
    """
    return F.cross_entropy(input, target)

def mse_loss(input, target):
    """
    Mean squared error loss function
    Formula: 1/N * sum((x - y)^2)
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        mse_loss: torch tensor of shape (1)
    """
    return F.mse_loss(input, target)

def l1_loss(input, target):
    """
    L1 loss function
    Formula: |x - y|
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        l1_loss: torch tensor of shape (1)
    """
    return F.l1_loss(input, target)

def hinge_loss(input, target):
    """
    Hinge loss function
    Formula: max(0, 1 - y * x)
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        hinge_loss: torch tensor of shape (1)
    """
    return F.hinge_embedding_loss(input, target)

def l2_loss(input, target):
    """
    L2 loss function
    Formula: (x - y)^2
    Input:
        input: torch tensor of shape (B, C, H, W)
        target: torch tensor of shape (B, C, H, W)
    Output:
        l2_loss: torch tensor of shape (1)
    """
    return F.mse_loss(input, target)




#function to train the model with accelerate

from accelerate import Accelerator
import torch


accelerator = Accelerator() #initialize the accelerator

def acc_train(model,config,train_loader, val_loader):
    
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    model.to(accelerator.device)
    model,optimizer,lr_scheduler ,train_loader,val_loader = accelerator.prepare(model,config.optimizer,config.lr_scheduler,train_loader,val_loader) #prepare the model, options, train_loader and val_loader for the accelerator
    
    model.train()
    for epoch in range(config.epochs):
        optimizer.zero_grad()
        for batch in train_loader:
            batch = {k: v.to(accelerator.device) for k, v in batch.items()}
            output = model(batch['image'])
            loss = config.loss_fn(output, batch['label'])
            accelerator.backward(loss)
            optimizer.step()
            lr_scheduler.step()
        
    model.eval()
    accurate = 0
    num_elems = 0
    for batch in val_loader:
        batch = {k: v.to(accelerator.device) for k, v in batch.items()}
        batch['image'] = batch['image'].to(device)
        batch['label'] = batch['label'].to(device)
        predictions = model(batch['image'])
        all_predictions, all_targets = accelerator.gather_for_metrics((predictions, batch['label']))
        num_elems += all_predictions.shape[0]
        accurate += all_predictions.long().sum() 
    eval_metric = accurate.item() / num_elems
    # Use accelerator.print to print only on the main process.
    accelerator.print(f"epoch {epoch}: {100 * eval_metric:.2f}")   